#!/usr/bin/env python
"""
pgtricks:0.9.2.dev
https://github.com/akaihola/pgtricks/blob/master/pgtricks/pg_dump_splitsort.py

Copyright (c) 2010, Antti Kaihola
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
       this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

    3. Neither the name of pgtricks nor the names of its contributors
       may be used to endorse or promote products derived from this software
       without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import functools
import io
import os
import re
from typing import Tuple, Union, cast, Pattern, Match, Optional, List, IO, Any

import sys

COPY_RE = re.compile(r'COPY .*? \(.*?\) FROM stdin;\n$')


def try_float(s1: str, s2: str) -> Union[Tuple[str, str], Tuple[float, float]]:
    if not s1 or not s2 or s1[0] not in '0123456789.-' or s2[0] not in '0123456789.-':
        # optimization
        return s1, s2
    try:
        return float(s1), float(s2)
    except ValueError:
        return s1, s2


def linecomp(l1: str, l2: str) -> int:
    p1 = l1.split('\t', 1)
    p2 = l2.split('\t', 1)
    v1, v2 = cast(Tuple[float, float], try_float(p1[0], p2[0]))
    result = (v1 > v2) - (v1 < v2)
    if not result and len(p1) == len(p2) == 2:
        return linecomp(p1[1], p2[1])
    return result

DATA_COMMENT_RE = re.compile('-- Data for Name: (?P<table>.*?); '
                             'Type: TABLE DATA; '
                             'Schema: (?P<schema>.*?);')
SEQUENCE_SET_RE = re.compile(r'-- Name: .+; Type: SEQUENCE SET; Schema: |'
                             r"SELECT pg_catalog\.setval\('")

class Matcher(object):
    def __init__(self) -> None:
        self._match: Optional[Match[str]] = None

    def match(self, pattern: Pattern[str], data: str) -> Optional[Match[str]]:
        self._match = pattern.match(data)
        return self._match

    def group(self, group1: str) -> str:
        if not self._match:
            raise ValueError('Pattern did not match')
        return self._match.group(group1)


def split_sql_file(sql_filepath: str) -> None:

    directory = os.path.dirname(sql_filepath)

    output: Optional[IO[str]] = None
    buf: List[str] = []

    def flush() -> None:
        cast(IO[str], output).writelines(buf)
        buf[:] = []

    def new_output(filename: str) -> IO[str]:
        if output:
            output.close()
        return open(os.path.join(directory, filename), 'w')

    copy_lines: Optional[List[str]] = None
    counter = 0
    output = new_output('0000_prologue.sql')
    matcher = Matcher()

    for line in open(sql_filepath):
        if copy_lines is None:
            if line in ('\n', '--\n'):
                buf.append(line)
            elif line.startswith('SET search_path = '):
                flush()
                buf.append(line)
            else:
                if matcher.match(DATA_COMMENT_RE, line):
                    counter += 1
                    output = new_output(
                        '{counter:04}_{schema}.{table}.sql'.format(
                            counter=counter,
                            schema=matcher.group('schema'),
                            table=matcher.group('table')))
                elif COPY_RE.match(line):
                    copy_lines = []
                elif SEQUENCE_SET_RE.match(line):
                    pass
                elif 1 <= counter < 9999:
                    counter = 9999
                    output = new_output('%04d_epilogue.sql' % counter)
                buf.append(line)
                flush()
        else:
            if line == '\\.\n':
                copy_lines.sort(key=functools.cmp_to_key(linecomp))
                buf.extend(copy_lines)
                buf.append(line)
                flush()
                copy_lines = None
            else:
                copy_lines.append(line)
    flush()


def main() -> None:
    split_sql_file(sys.argv[1])


if __name__ == '__main__':
    main()


import os
from flask import Flask
from flask_migrate import Migrate

from ibdb.models import db
from ibdb.routes import blueprints

app = Flask(__name__, template_folder='./templates')
app.config.from_object(os.environ.get('APP_SETTINGS', 'ibdb.settings.FlaskConfig'))
db.init_app(app)
app.config['db'] = db

migrate = Migrate()
migrate.init_app(app, db)

for bp in blueprints:
    app.register_blueprint(bp)

if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0')



from flask import current_app, Blueprint, jsonify, request, redirect, render_template

from ibdb.models import Episode, Reference, Show

episode_bp = Blueprint('episode_bp', __name__, template_folder='templates')


@episode_bp.record
def record(state):
    db = state.app.config.get("db")
    if db is None:
        raise Exception("This blueprint expects you to provide database access through db")


@episode_bp.route('/episodes')
def episodes():
    episodes = Episode.query.order_by(Episode.published_date.desc(), Episode.id).all()
    if request.args.get('format', None) == 'json':
        return jsonify([e.serialize() for e in episodes])
    else:
        references = Reference.query.order_by(Reference.name, Reference.id).all()
        shows = Show.query.order_by(Show.id).all()
        return render_template('episode/list_edit.html', episodes=episodes, references=references, shows=shows)


@episode_bp.route('/episodes', methods=['POST'])
def episode_new():
    new = Episode(
        id=request.form.get("episode_id"),
        name=request.form.get("name"),
        youtube_link=request.form.get("youtube_link"),
        official_link=request.form.get("official_link"),
        image_link=request.form.get("image_link"),
        published_date=request.form.get("published_date"),
        show_id=request.form.get("show_id"),
    )
    db = current_app.config['db']
    db.session.add(new)
    db.session.commit()
    return redirect('/episodes#' + str(new.id))


@episode_bp.route('/episode/<id>', methods=['GET'])
def episode_by(id):
    ep = Episode.query.filter_by(id=id).first_or_404()
    return jsonify(ep.serialize())


@episode_bp.route('/episode/<id>', methods=['POST'])
def episode_update(id):
    Episode.query.filter_by(id=id).update(request.form.to_dict())
    db = current_app.config['db']
    db.session.commit()
    redirect('/episodes#' + str(id))


@episode_bp.route('/episode/<id>/inspired_by', methods=['POST'])
def episode_inspired_by(id):
    ep = Episode.query.get(id)
    ref = Reference.query.get(request.form.get("inspiration_reference_id"))
    ep.inspired_by.append(ref)
    db = current_app.config['db']
    db.session.add(ep)
    db.session.commit()
    return redirect('/episodes#' + str(ep.id))



from .base.base import base_bp
from .episode.episode import episode_bp
from .guest.guest import guest_bp
from .recipe.recipe import recipe_bp
from .reference.reference import reference_bp
from .show.show import show_bp

blueprints = [
    base_bp,
    episode_bp,
    guest_bp,
    recipe_bp,
    reference_bp,
    show_bp,
]



from flask import jsonify, request, Blueprint, current_app

from ibdb.models import Show

show_bp = Blueprint('show_bp', __name__, template_folder='templates')


@show_bp.record
def record(state):
    db = state.app.config.get("db")
    if db is None:
        raise Exception("This blueprint expects you to provide database access through db")


@show_bp.route('/shows')
def shows():
    shows = Show.query.order_by(Show.id).all()
    return jsonify([s.serialize() for s in shows])


@show_bp.route('/show/<id>')
def show_by(id):
    s = Show.query.filter_by(id=id).first_or_404()
    return jsonify(s.serialize())


@show_bp.route('/show/<id>', methods=['POST'])
def show_update(id):
    Show.query.filter_by(id=id).update(request.form.to_dict())
    db = current_app.config['db']
    db.session.commit()
    updated = Show.query.filter_by(id=id).first()
    return jsonify(updated.serialize())



import os

import click

from ibdb.api import app
from ibdb.populate_babish_json import populate_babish_json
from ibdb.populate_db import populate_db
from ibdb.export import export as sync_export


@app.cli.command('api')
def api():
    app.run(debug=True, host='0.0.0.0')


@app.cli.group('sync', chain=True)
def sync():
    pass


@sync.command('update')
def update():
    populate_db()


@sync.command('export')
@click.option('--format', '-f', multiple=True, default=["json", "sql"], type=click.Choice(['json', 'sql', 'babish-json']),
              help='Output formats (default: json,sql)')
def export(format):
    sync_export(json_='json' in format, sql='sql' in format)
    if 'babish-json' in format:
        populate_babish_json()


def main():
    os.environ['FLASK_APP'] = 'ibdb.api'
    return app.cli.main()



# coding: utf-8
from sqlalchemy import Column, ForeignKey, Integer, String, Table, Text, text, Date
from sqlalchemy.orm import relationship, backref
from flask_sqlalchemy import SQLAlchemy
from flask import request

from ibdb.recipe_parser import RecipeParser

db = SQLAlchemy()
Base = db.Model
metadata = Base.metadata


class Guest(Base):
    __tablename__ = 'guest'

    id = Column(Integer, primary_key=True, server_default=text("nextval('guest_id_seq'::regclass)"))
    name = Column(String, nullable=False, unique=True)
    image_link = Column(String)

    def serialize(self, related=True):
        return {
            'self': f'{request.host_url[:-1]}/guest/{self.id}',
            'guest_id': self.id,
            'name': self.name,
            'image_link': self.image_link,
            **({
                'appearances': [
                    e.serialize(False)
                    for e in self.appearances
                ]
            } if related else {})
        }


class Reference(Base):
    __tablename__ = 'reference'

    id = Column(Integer, primary_key=True, server_default=text("nextval('reference_id_seq'::regclass)"))
    type = Column(String)
    name = Column(String, unique=True)
    description = Column(Text)
    external_link = Column(String)
    image_link = Column(String)

    def serialize(self, related=True):
        return {
            'self': f'{request.host_url[:-1]}/reference/{self.id}',
            'reference_id': self.id,
            'type': self.type,
            'name': self.name,
            'description': self.description,
            'external_link': self.external_link,
            'image_link': self.image_link,
            **({
                'episodes_inspired': [
                    e.serialize(False)
                    for e in self.episodes_inspired
                ]
            } if related else {})
        }


class Show(Base):
    __tablename__ = 'show'

    id = Column(Integer, primary_key=True, server_default=text("nextval('show_id_seq'::regclass)"))
    name = Column(String, nullable=False, unique=True)
    image_link = Column(String)

    def serialize(self, related=True):
        return {
            'self': f'{request.host_url[:-1]}/show/{self.id}',
            'show_id': self.id,
            'name': self.name,
            'image_link': self.image_link,
            **({
                'episodes': [
                    e.serialize(False)
                    for e in self.episodes
                ]
            } if related else {})
        }


class Episode(Base):
    __tablename__ = 'episode'

    id = Column(String, primary_key=True)
    name = Column(String, nullable=False, unique=True)
    youtube_link = Column(String, nullable=False)
    official_link = Column(String, nullable=False)
    image_link = Column(String)
    published_date = Column(Date())
    show_id = Column(ForeignKey('show.id'), nullable=False)

    _backref_order_by = 'Episode.published_date.desc(),Episode.id'

    show = relationship('Show',
                        uselist=False,
                        order_by='Show.id',
                        backref=backref('episodes',
                                        order_by=_backref_order_by))
    guests = relationship('Guest',
                          secondary='guest_appearances',
                          order_by='Guest.name,Guest.id',
                          backref=backref('appearances',
                                          order_by=_backref_order_by))
    inspired_by = relationship('Reference',
                               secondary='episode_inspired_by',
                               order_by='Reference.name,Reference.id',
                               backref=backref('episodes_inspired',
                                               order_by=_backref_order_by))
    recipes = relationship('Recipe', order_by='Recipe.name,Recipe.id',
                           backref=backref('episode',
                                           uselist=False,
                                           order_by=_backref_order_by))

    def serialize(self, related=True):
        return {
            'self': f'{request.host_url[:-1]}/episode/{self.id}',
            'episode_id': self.id,
            'name': self.name,
            'published_date': self.published_date.isoformat(),
            'youtube_link': self.youtube_link,
            'official_link': self.official_link,
            'image_link': self.image_link,
            **({
                'related': {
                    'show': self.show.serialize(False),
                    'guests': [
                        g.serialize(False)
                        for g in self.guests
                    ],
                    'inspired_by': [
                        r.serialize(False)
                        for r in self.inspired_by
                    ],
                    'recipes': [
                        r.serialize(False)
                        for r in self.recipes
                    ],
                }
            } if related else {})
        }


t_episode_inspired_by = Table(
    'episode_inspired_by', metadata,
    Column('episode_id', ForeignKey('episode.id'), primary_key=True, nullable=False),
    Column('reference_id', ForeignKey('reference.id'), primary_key=True, nullable=False),
)


t_guest_appearances = Table(
    'guest_appearances', metadata,
    Column('episode_id', ForeignKey('episode.id'), primary_key=True, nullable=False),
    Column('guest_id', ForeignKey('guest.id'), primary_key=True, nullable=False),
)


class Recipe(Base):
    __tablename__ = 'recipe'

    id = Column(Integer, primary_key=True, server_default=text("nextval('recipe_id_seq'::regclass)"))
    name = Column(String, nullable=False)
    image_link = Column(String)
    raw_ingredient_list = Column(Text, nullable=False)
    raw_procedure = Column(Text, nullable=False)
    episode_id = Column(ForeignKey('episode.id'), nullable=False)

    def serialize(self, related=True):
        return {
            'self': f'{request.host_url[:-1]}/recipe/{self.id}',
            'recipe_id': self.id,
            'name': self.name,
            'image_link': self.image_link,
            'raw_ingredient_list': self.raw_ingredient_list,
            'raw_procedure': self.raw_procedure,
            'ingredient_list': self.ingredient_list(),
            **({
                'source': self.episode.serialize(False),
            } if related else {})
        }

    def ingredient_list(self):
        return [
            RecipeParser.parse_ingredient(i)
            for i in self.raw_ingredient_list.splitlines()
            if 'For the ' not in i
        ]


'''
Helper methods for models
'''
# TODO: refactor to use ORM


def upsert_episode(session, episode):
    # test if we will be creating this
    result = session.execute(
        """
            SELECT id FROM episode WHERE id = :id
        """,
        {k: episode[k] for k in ('id',)}
    )
    created = True
    if result.rowcount > 0:
        created = False

    # upsert
    session.execute(
        """
            INSERT INTO episode (id, name, youtube_link, official_link, image_link, published_date, show_id)
            VALUES (:id, :name, :youtube_link, :official_link, :image_link, :published_date, :show_id)
            ON CONFLICT (id)
            DO UPDATE SET (name, youtube_link, official_link, image_link, published_date, show_id)
            = (EXCLUDED.name, EXCLUDED.youtube_link, EXCLUDED.official_link, EXCLUDED.image_link, EXCLUDED.published_date, EXCLUDED.show_id)
        """,
        {k: episode[k] for k in ('id', 'name', 'youtube_link', 'official_link', 'image_link', 'published_date', 'show_id')})
    return episode['id'], created


def upsert_reference(session, reference):
    result = session.execute(
        """
            SELECT id FROM reference WHERE name = :name
        """,
        {k: reference[k] for k in ('name',)}
    )
    if result.rowcount == 0:
        result = session.execute(
            """
                INSERT INTO reference (name)
                VALUES (:name)
                RETURNING id
            """,
            {k: reference[k] for k in ('name',)}
        )
    reference_id = result.fetchone()[0]
    return reference_id


def upsert_episode_inspired_by(session, episode_id, reference_id):
    session.execute(
        """
            INSERT INTO episode_inspired_by (episode_id, reference_id)
            VALUES (:episode_id, :reference_id)
            ON CONFLICT (episode_id, reference_id)
            DO NOTHING
        """,
        {'episode_id': episode_id, 'reference_id': reference_id}
    )


def upsert_guest(session, guest):
    result = session.execute(
        """
            SELECT id FROM guest WHERE name = :name
        """,
        {k: guest[k] for k in ('name',)}
    )
    if result.rowcount == 0:
        result = session.execute(
            """
                INSERT INTO guest (name)
                VALUES (:name)
                RETURNING id
            """,
            {k: guest[k] for k in ('name',)}
        )
    reference_id = result.fetchone()[0]
    return reference_id


def upsert_guest_appearance(session, episode_id, guest_id):
    session.execute(
        """
            INSERT INTO guest_appearances (episode_id, guest_id)
            VALUES (:episode_id, :guest_id)
            ON CONFLICT (episode_id, guest_id)
            DO NOTHING
        """,
        {'episode_id': episode_id, 'guest_id': guest_id}
    )


def count_episode_recipes(session, episode_id):
    result = session.execute(
        """
            SELECT id FROM recipe WHERE episode_id = :episode_id
        """,
        {'episode_id': episode_id}
    )
    return result.rowcount


def upsert_recipe(session, recipe):
    result = session.execute(
        """
            SELECT id FROM recipe WHERE episode_id = :episode_id AND name = :name
        """,
        {k: recipe[k] for k in ('episode_id', 'name')}
    )
    if result.rowcount == 0:
        result = session.execute(
            """
                INSERT INTO recipe (episode_id, name, image_link, raw_ingredient_list, raw_procedure)
                VALUES (:episode_id, :name, :image_link, :raw_ingredient_list, :raw_procedure)
                RETURNING id
            """,
            {k: recipe[k] for k in ('episode_id', 'name', 'image_link', 'raw_ingredient_list', 'raw_procedure')}
        )
    recipe_id = result.fetchone()[0]
    return recipe_id



import re

from collections import namedtuple
from fractions import Fraction
from string import capwords, whitespace
from unicodedata import numeric


class RecipeParser(object):

    Ingredient = namedtuple('Ingredient', 'qty unit name raw')
    units_pattern = r'(?:(\s?mg|g|kg|ml|L|oz|ounce|tbsp|Tbsp|tablespoon|tsp|teaspoon|cup|lb|pound|small|medium|large|whole|half)?(?:s|es)?\.?\b)'

    full_pattern = r'^(?:([-\.\/\s0-9\u2150-\u215E\u00BC-\u00BE]+)?{UNITS_PATTERN})?(?:.*\sof\s)?\s?(.+?)(?:,|$)'.format(
        UNITS_PATTERN=units_pattern)

    pattern = re.compile(full_pattern, flags=re.UNICODE)

    # https://en.wikipedia.org/wiki/Cooking_weights_and_measures#United_States_measures
    measures = {
        'drop': {'abrv': 'dr gt gtt', 'oz': 1.0 / 576},
        'smidgen': {'abrv': 'smdg smi', 'oz': 1.0 / 256},
        'pinch': {'abrv': 'pn', 'oz': 1.0 / 128},
        'dash': {'abrv': 'ds', 'oz': 1.0 / 64},
        'saltspoon': {'abrv': 'ssp scruple', 'oz': 1.0 / 32},
        'coffeespoon': {'abrv': 'csp', 'oz': 1.0 / 16},
        'dram': {'abrv': 'dr', 'oz': 1.0 / 8},
        'teaspoon': {'abrv': 'tsp t', 'oz': 1.0 / 6},
        'tablespoon': {'abrv': 'Tbsp T', 'oz': 1.0 / 2},
        'ounce': {'abrv': 'oz fl.oz', 'oz': 1.0},
        'wineglass': {'abrv': 'wgf', 'oz': 2.0},
        'teacup': {'abrv': 'tcf gill', 'oz': 4.0},
        'cup': {'abrv': 'C', 'oz': 8.0},
        'pint': {'abrv': 'pt', 'oz': 16.0},
        'quart': {'abrv': 'qt', 'oz': 32.0},
        'pottle': {'abrv': 'pot', 'oz': 64.0},
        'gallon': {'abrv': 'gal', 'oz': 128.0},
        'pound': {'abrv': 'lbs', 'oz': 16.0},
        'gram': {'abrv': 'g', 'oz': 0.035274},
        'kilogram': {'abrv': 'kg', 'oz': 35.274},
    }

    @classmethod
    def parse_ingredient(cls, i):
        if not isinstance(i, str):
            if i.string is None:
                # multiple tags in child,
                # bs4 gets confused per https://www.crummy.com/software/BeautifulSoup/bs4/doc/#string
                s = i.get_text().strip()
            else:
                s = i.string.strip()
        else:
            s = i
        raw = s.replace('\xa0', '').strip()

        clean = re.sub(r'\(.+?\)', '', raw).replace('’', "'")

        parsed = cls.pattern.match(clean)
        if parsed:
            qty, unit, name = parsed.groups()
        else:
            print("WARN: Unable to parse ingredient '{}'".format(raw))
            return cls.Ingredient(None, None, None, raw)

        qty = cls.normalize_qty(qty)

        unit = cls.normalize_unit(unit)

        name = cls.normalize_name(name)

        return cls.Ingredient(qty, unit, name, raw)

    @classmethod
    def normalize_qty(cls, qty):
        if qty is not None:
            qty = qty.strip()

            if len(qty) == 0:
                qty = None
            elif len(qty) == 1:
                qty = numeric(qty)
            else:
                try:
                    if '/' in qty:
                        # 2 1/2
                        qty = float(sum(Fraction(s) for s in qty.split()))
                    elif qty[-1].isdigit():
                        # normal number, ending in [0-9]
                        qty = float(qty)
                    else:
                        # Assume the last character is a vulgar fraction
                        qty = float(qty[:-1]) + numeric(qty[-1])
                except ValueError:
                    pass  # let it be a string
        return qty

    @classmethod
    def normalize_unit(cls, unit):
        if unit is not None:
            for u, d in cls.measures.items():
                if unit.lower() == u:
                    return u
                if unit in d['abrv'].split(' '):
                    return u
        return unit

    @classmethod
    def normalize_name(cls, name):
        return capwords(name) \
            .replace('Whole ', '').replace('Half ', '') \
            .replace('Hot ', '').replace('Warm ', '').replace('Cold ', '') \
            .strip(whitespace + ',.')



import json
import logging
import os
import subprocess
from pathlib import Path

import flask

import ibdb.api
from ibdb import settings
from vendor.pg_dump_splitsort import split_sql_file

IBDB_URL = settings.IBDB_URL
IBDB_SQLFILE = settings.IBDB_SQLFILE
OUTPUT_DIR = settings.OUTPUT_DIR

DB_USERNAME = settings.DB_USERNAME
DB_PASSWORD = settings.DB_PASSWORD
DB_NAME = settings.DB_NAME
DB_HOSTNAME = settings.DB_HOSTNAME
DB_PORT = settings.DB_PORT


def write_response(route, to=None, **kwargs):
    logging.info(f'Writing {route} to {to}')
    kwargs['base_url'] = kwargs.get('base_url') or IBDB_URL
    with ibdb.api.app.test_client() as c:
        r: flask.Response = c.get(route, **kwargs)
        with open(to, 'w', encoding='utf8') as o:
            json.dump(r.json, o, indent=2)


def write_datasets():
    logging.info(f'Writing datasets to {OUTPUT_DIR}')
    write_response('/episodes?format=json', to=os.path.join(OUTPUT_DIR, 'ibdb.episodes.json'))
    write_response('/guests?format=json', to=os.path.join(OUTPUT_DIR, 'ibdb.guests.json'))
    write_response('/recipes?format=json', to=os.path.join(OUTPUT_DIR, 'ibdb.recipes.json'))
    write_response('/references?format=json', to=os.path.join(OUTPUT_DIR, 'ibdb.references.json'))
    write_response('/shows?format=json', to=os.path.join(OUTPUT_DIR, 'ibdb.shows.json'))


def cleanup_sqlfile(sql_filepath):
    base_dir = os.path.dirname(sql_filepath)
    split_sql_file(sql_filepath)
    temp_sql_files = sorted(Path(base_dir).glob('[0-9]*.sql'))
    if len(temp_sql_files) > 0:
        with open(sql_filepath, 'w') as f:
            subprocess.run(['cat', *temp_sql_files], stdout=f)
        subprocess.run(['rm', *temp_sql_files])
    else:
        logging.error('pg_dump_splitsort did not produce the files expected')


def db_dump():
    logging.info(f'Writing database to {IBDB_SQLFILE}')
    subprocess.run(f'''
        PGPASSWORD={DB_PASSWORD} PGUSER={DB_USERNAME} PGHOST={DB_HOSTNAME} PGPORT={DB_PORT} \
        pg_dump --dbname={DB_NAME} --create --file={IBDB_SQLFILE}''', shell=True)
    cleanup_sqlfile(IBDB_SQLFILE)


def export(json_=True, sql=True):
    if json_:
        write_datasets()
    if sql:
        db_dump()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')

    export()



from flask import jsonify, request, render_template, redirect, Blueprint, current_app

from ibdb.models import Recipe, Episode
from ibdb.utils import normalize_raw_list

recipe_bp = Blueprint('recipe_bp', __name__, template_folder='templates')


@recipe_bp.record
def record(state):
    db = state.app.config.get("db")
    if db is None:
        raise Exception("This blueprint expects you to provide database access through db")


@recipe_bp.route('/recipes')
def recipes():
    recipes = Recipe.query.order_by(Recipe.name, Recipe.id).all()
    if request.args.get('format', None) == 'json':
        return jsonify([r.serialize() for r in recipes])
    else:
        episodes = Episode.query.order_by(Episode.published_date.desc(), Episode.id).all()
        return render_template('recipe/list_edit.html', recipes=recipes, episodes=episodes)


@recipe_bp.route('/recipes', methods=['POST'])
def recipe_new():
    new = Recipe(
        name=request.form.get("name"),
        image_link=request.form.get("image_link"),
        raw_ingredient_list=normalize_raw_list(
            request.form.get("raw_ingredient_list")),
        raw_procedure=normalize_raw_list(
            request.form.get("raw_procedure")),
        episode_id=request.form.get("source_episode_id"),
    )
    db = current_app.config['db']
    db.session.add(new)
    db.session.commit()
    return redirect('/recipes#r' + str(new.id))


@recipe_bp.route('/recipe/<id>')
def recipe_by(id):
    r = Recipe.query.filter_by(id=id).first_or_404()
    return jsonify(r.serialize())


@recipe_bp.route('/recipe/<id>', methods=['POST'])
def recipe_update(id):
    Recipe.query.filter_by(id=id).update(request.form.to_dict())
    db = current_app.config['db']
    db.session.commit()
    return redirect('/recipes#r' + str(id))



import os

IBDB_URL = os.environ.get('IDDB_URL', 'ibdb://data-with-babish.jklewa.github.com')
IBDB_SQLFILE = os.environ.get('IBDB_SQLFILE', 'ibdb/db_dump.sql')
OUTPUT_DIR = os.environ.get('OUTPUT_DIR', 'datasets/')

DB_USERNAME = os.environ.get('POSTGRES_USERNAME', 'postgres')
DB_PASSWORD = os.environ.get('POSTGRES_PASSWORD', 'postgres')
DB_NAME = os.environ.get('POSTGRES_DBNAME', 'babish_db')
DB_HOSTNAME = os.environ.get('POSTGRES_HOSTNAME', 'db')
DB_PORT = int(os.environ.get('POSTGRES_PORT', '5432'))


class FlaskConfig(object):
    DEBUG = False
    TESTING = False
    JSON_SORT_KEYS = False
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    SQLALCHEMY_DATABASE_URI = f"postgresql+psycopg2://{DB_USERNAME}:{DB_PASSWORD}@{DB_HOSTNAME}:{DB_PORT}/{DB_NAME}"



from flask import jsonify, request, render_template, redirect, current_app, Blueprint

from ibdb.models import Reference, Episode

reference_bp = Blueprint('reference_bp', __name__, template_folder='templates')


@reference_bp.record
def record(state):
    db = state.app.config.get("db")
    if db is None:
        raise Exception("This blueprint expects you to provide database access through db")


@reference_bp.route('/references')
def references():
    references = Reference.query.order_by(Reference.name, Reference.id).all()
    if request.args.get('format', None) == 'json':
        return jsonify([r.serialize() for r in references])
    else:
        episodes = Episode.query.order_by(Episode.published_date.desc(), Episode.id).all()
        return render_template('reference/list_edit.html', references=references, episodes=episodes)


@reference_bp.route('/references', methods=['POST'])
def reference_new():
    new = Reference(
        type=request.form.get("type"),
        name=request.form.get("name"),
        image_link=request.form.get("image_link"),
        description=request.form.get("description"),
        external_link=request.form.get("external_link"),
    )
    inspired = Episode.query.get(
        request.form.get("inspired_episode_id")
    )
    if inspired:
        new.episodes_inspired.append(inspired)
    db = current_app.config['db']
    db.session.add(new)
    db.session.commit()
    return redirect('/references#r' + str(new.id))


@reference_bp.route('/reference/<id>/episodes_inspired', methods=['POST'])
def reference_episodes_inspired(id):
    ref = Reference.query.get(id)
    inspired = Episode.query.get(request.form.get("inspired_episode_id"))
    ref.episodes_inspired.append(inspired)
    db = current_app.config['db']
    db.session.add(ref)
    db.session.commit()
    return redirect('/references#r' + str(ref.id))


@reference_bp.route('/reference/<id>')
def reference_by(id):
    r = Reference.query.filter_by(id=id).first_or_404()
    return jsonify(r.serialize())


@reference_bp.route('/reference/<id>', methods=['POST'])
def reference_update(id):
    Reference.query.filter_by(id=id).update(request.form.to_dict())
    db = current_app.config['db']
    db.session.commit()
    return redirect('/references#r' + str(id))



# coding: utf-8

import bs4
import json
import re
import os
import logging
import string

from ibdb import settings
from ibdb.populate_db import fetch_binging_episode_list
from ibdb.recipe_parser import RecipeParser
from ibdb.utils import timeit, Stats

OUTPUT_DIR = settings.OUTPUT_DIR


class PopulateBabishJSON:

    def __init__(self):
        self.episodes = []
        self.stats = Stats({
            'episodes': 0,
            'bytes_written': 0,
        })

    @timeit
    def fetch_episode_list(self):
        # Fetch the current list of BWB episodes

        self.episodes = [e for e in fetch_binging_episode_list()]
        self.stats.episodes = len(self.episodes)

    @timeit
    def generate_babish_json(self):
        # Parse all episodes into babish.json

        episodes = []
        for episode in self.episodes:
            soup = bs4.BeautifulSoup(episode['body'], 'html.parser')

            ep = {
                'episode_name': episode['name'],
                'episode_link': episode['official_link'],
                'youtube_link': episode['youtube_link'],
                'published': episode['published_date'],
                'recipes': []
            }

            recipe_locations = soup.find_all(['h1', 'h2', 'h3', 'h4', 'h5'], string=re.compile('Ingredients'))

            for loc in recipe_locations:
                ingredient_lists = loc.parent.find_all(['ul'])

                for iloc in ingredient_lists:
                    # TODO: This logic is messy and brittle. Refactor it.

                    # Get method name, usually right before the list itself
                    method_name = iloc.find_previous_sibling(['p', 'h2', 'h3'])
                    if method_name is not None:
                        method_name = method_name.get_text().replace('Ingredients', '').strip(string.whitespace + '\u00a0:,')

                    if method_name is None or method_name == '':
                        # Look for the method name above the recipe instead
                        h = iloc.parent.find_next(['h1', 'h2'])
                        if h:
                            method_name = h.get_text().strip()
                            method_name = re.sub('^(Method)', '', method_name).strip(string.whitespace + '\u00a0:,')
                        else:
                            # No luck, fall back to default
                            method_name = ''

                    # Default to the episode name
                    if method_name == '':
                        method_name = episode['name']

                    def flatten(loc):
                        out = []
                        for i in loc.children:
                            # flatten nested lists
                            if isinstance(i, bs4.element.Tag) and i.select('ul,ol'):
                                for i2 in i.select('ul,ol'):
                                    out += flatten(i2)
                            else:
                                out.append(i)
                        return out

                    # Convert the ul to parsed Ingredients
                    try:
                        raw_ingredients = flatten(iloc)
                        ingredients = [RecipeParser.parse_ingredient(i) for i in raw_ingredients]
                    except Exception:
                        print("ERROR: Failed to parse ingredients from ep: {0} method: {1} raw_list: {2}".format(
                            episode['name'],
                            method_name,
                            list(iloc.children)
                        ))
                        raise

                    if len(ingredients) == 0:
                        print(
                            "WARN: Could not find ingredients for {0} (Episode {1})".format(
                                method_name, episode['name']))

                    recipe = {
                        'method': method_name,
                        'ingredients': ingredients,
                    }

                    ep['recipes'].append(recipe)

            episodes.append(ep)

        with open(os.path.join(OUTPUT_DIR, 'babish.json'), 'w', encoding='utf8') as f:
            json.dump(episodes, f, indent=2, ensure_ascii=False)

        self.stats.bytes_written = os.path.getsize(os.path.join(OUTPUT_DIR, 'babish.json'))

        logging.info('Wrote babish.json successfully, %s bytes written', self.stats.bytes_written)

    def show_stats(self):
        print(self.stats)


def populate_babish_json():
    BABISH = PopulateBabishJSON()

    BABISH.fetch_episode_list()
    BABISH.generate_babish_json()

    BABISH.show_stats()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')

    populate_babish_json()



import logging
import re
import time
from datetime import datetime


def normalize_raw_list(raw):
    return '\n'.join(
        line
        for line in (
            re.sub(r'^(\*\s+|[0-9]+\.\s+)', '', line.strip())
            for line in raw.splitlines()
        )
        if line
    )


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            logging.info('Time taken in %s  %2.2f ms' %
                         (method.__name__, (te - ts) * 1000))
        return result
    return timed


class Stats(object):

    def __init__(self, d):
        self.__dict__ = d

    def __repr__(self):
        return 'Stats:\n{0}'.format('\n'.join(' {0}: {1}'.format(stat, val) for stat, val in self.__dict__.items()))


def timestamp_to_date(ts_in_milli):
    if not ts_in_milli:
        return None
    return datetime.fromtimestamp(ts_in_milli / 1000).strftime('%Y-%m-%d')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "beautifulsoup4",
    "click",
    "flask",
    "flask-migrate",
    "flask-sqlalchemy",
    "furl",
    "html5lib",
    "measurement",
    "psycopg2-binary",
    "pyfunctional",
    "requests",
]

setuptools.setup(
    name="babish-db",
    version="0.0.1",
    license='GPLv3+',
    author="jklewa",
    author_email="jklewa@gmail.com",
    description="Data with Babish DB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jklewa/data-with-babish",
    scripts=[
        "ibdb/bin/ibdb",
    ],
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import logging
import html

from ibdb.api import db
from ibdb.extract import fetch_paginated_seq, extract_inspired_by, \
    add_youtube_resources, extract_guests, extract_recipes
from ibdb.models import upsert_episode, upsert_reference, upsert_episode_inspired_by, upsert_guest, \
    upsert_guest_appearance, upsert_recipe, count_episode_recipes
from ibdb.utils import timestamp_to_date

logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')


def fetch_binging_episode_list():
    # Fetch the current list of Binging with Babish episodes

    base_url = 'https://www.bingingwithbabish.com'
    route = '/recipes'

    raw_episode_list = fetch_paginated_seq(base_url, route)

    episodes = raw_episode_list.map(add_youtube_resources).map(lambda raw: {
        'id': raw.get('urlId', None).split('/')[-1],
        'show_id': 1,
        'name': html.unescape(raw.get('title', '')),
        'official_link': base_url + raw.get('fullUrl', ''),
        'youtube_link': raw['youtube_link'],
        'image_link': raw['youtube_image_link'] or raw.get('assetUrl', None),
        'published_date': timestamp_to_date(raw.get('publishOn', None)),
        'body': raw['body'],
    })
    episode_count = sum(1 for x in episodes)

    logging.info('Fetched episode list, %s episodes', episode_count)
    return reversed(episodes.to_list())


def fetch_basics_episode_list():
    # Fetch the current list of Basics with Babish episodes

    base_url = 'https://basicswithbabish.co'
    route = '/basicsepisodes'

    raw_episode_list = fetch_paginated_seq(base_url, route)

    episodes = raw_episode_list.map(add_youtube_resources).map(lambda raw: {
        'id': raw.get('urlId', None).split('/')[-1],
        'show_id': 2,
        'name': html.unescape(raw.get('title', '')),
        'official_link': base_url + raw.get('fullUrl', ''),
        'youtube_link': raw['youtube_link'],
        'image_link': raw['youtube_image_link'] or raw.get('assetUrl', None),
        'published_date': timestamp_to_date(raw.get('publishOn', None)),
        'body': raw['body'],
    })
    episode_count = sum(1 for x in episodes)

    logging.info('Fetched episode list, %s episodes', episode_count)
    return reversed(episodes.to_list())


def populate_binging(session):
    episodes = fetch_binging_episode_list()

    for ep in episodes:
        ep['id'], created = upsert_episode(session, ep)
        if created:
            print("NEW {published_date} | {name}".format(**ep))

        inspiration_list = extract_inspired_by(ep['name'])
        for inspired_by in inspiration_list:
            ref = {
                'name': inspired_by,
            }
            ref['id'] = upsert_reference(session, ref)
            upsert_episode_inspired_by(session, ep['id'], ref['id'])

        guest_list = extract_guests(ep['name'])
        for guest in guest_list:
            guest = {
                'name': guest,
            }
            guest['id'] = upsert_guest(session, guest)
            upsert_guest_appearance(session, ep['id'], guest['id'])

        if created or count_episode_recipes(session, ep['id']) == 0:
            recipes = extract_recipes(ep)
            if len(recipes) > 0:
                if not created:
                    print("    {published_date} | {name}".format(**ep))
                print(f'                 {len(recipes)} new recipes')
            for recipe in recipes:
                upsert_recipe(session, recipe)


def populate_basics(session):
    episodes = fetch_basics_episode_list()
    for ep in episodes:
        ep['id'], created = upsert_episode(session, ep)
        if created:
            print("NEW {published_date} | {name}".format(**ep))

        guest_list = extract_guests(ep['name'])
        for guest in guest_list:
            guest = {
                'name': guest,
            }
            guest['id'] = upsert_guest(session, guest)
            upsert_guest_appearance(session, ep['id'], guest['id'])

        if created or count_episode_recipes(session, ep['id']) == 0:
            recipes = extract_recipes(ep)
            if len(recipes) > 0:
                if not created:
                    print("    {published_date} | {name}".format(**ep))
                print(f'                 {len(recipes)} new recipes')
            for recipe in recipes:
                upsert_recipe(session, recipe)


def populate_db():
    session = db.session()

    populate_binging(session)
    populate_basics(session)

    session.commit()
    print('Done')


if __name__ == '__main__':
    populate_db()



from flask import jsonify, request, render_template, redirect, Blueprint, current_app

from ibdb.models import Guest, Episode

guest_bp = Blueprint('guest_bp', __name__, template_folder='templates')


@guest_bp.record
def record(state):
    db = state.app.config.get("db")
    if db is None:
        raise Exception("This blueprint expects you to provide database access through db")


@guest_bp.route('/guests')
def guests():
    guests = Guest.query.order_by(Guest.name).all()
    if request.args.get('format', None) == 'json':
        return jsonify([g.serialize() for g in guests])
    else:
        episodes = Episode.query.order_by(Episode.published_date.desc(), Episode.id).all()
        return render_template('guest/list_edit.html', guests=guests, episodes=episodes)


@guest_bp.route('/guests', methods=['POST'])
def guest_new():
    new = Guest(
        name=request.form.get("name"),
        image_link=request.form.get("image_link"),
    )
    appearance = Episode.query.get(
        request.form.get("appearance_episode_id")
    )
    if appearance:
        new.appearances.append(appearance)
    db = current_app.config['db']
    db.session.add(new)
    db.session.commit()
    return redirect('/guests#g' + str(new.id))


@guest_bp.route('/guest/<id>/appearances', methods=['POST'])
def guest_appearances(id):
    guest = Guest.query.get(id)
    appearance = Episode.query.get(request.form.get("appearance_episode_id"))
    guest.appearances.append(appearance)
    db = current_app.config['db']
    db.session.add(guest)
    db.session.commit()
    return redirect('/guests#g' + str(guest.id))


@guest_bp.route('/guest/<id>')
def guest_by(id):
    g = Guest.query.filter_by(id=id).first_or_404()
    return jsonify(g.serialize())


@guest_bp.route('/guest/<id>', methods=['POST'])
def guest_update(id):
    Guest.query.filter_by(id=id).update(request.form.to_dict())
    db = current_app.config['db']
    db.session.commit()
    return redirect('/guests#g' + str(id))



import json
import logging
import re
import string

import bs4
import requests
from functional import seq
from furl import furl

from ibdb.recipe_parser import RecipeParser

logging.basicConfig(level=logging.INFO, format='%(levelname)s: %(message)s')


def extract_inspired_by(name):
    for sub in ('"', '“', '...', 'Real Sample', 'Part II', 'Part I', 'Lots of Stuff', 'Duck Carbonara',
                'Cocktail Special', 'Subscriber Special'):
        name = name.replace(sub, ' ')
    inspired_by = re.match(r'.*(?:inspired by|from) ([^(|["“!;:]{2,}?)([(|["“!;:]|with|feat\.|$)', name, re.IGNORECASE & re.UNICODE)
    if inspired_by:
        inspiration_list = [i.strip() for i in re.split(r'\s(?:and(?! the| Rec))\s', inspired_by[1])]

        additional = re.match(r'.*\((?:and|plus) ([^\s]+.+)\)', name, re.IGNORECASE & re.UNICODE)
        if additional:
            inspiration_list.extend(i.strip() for i in re.split(r'\s(?:and(?! the))\s', additional[1]))

        return inspiration_list

    special = re.match(r'(.+) (?:Special|Volume I|Burger Cookoff$|Thanksgiving$)', name, re.IGNORECASE & re.UNICODE)
    if special:
        return [special[1].strip()]
    return []


def extract_guests(name):
    for sub in ('"', '“', '...', 'Homemade Ciabatta', 'with Babish', 'Wild Game inspired by Game of Thrones'):
        name = name.replace(sub, ' ')
    guest = re.match(r'.*(?:feat\.|featuring|with) ([^\s][^)]+)(?:$|\))', name, re.IGNORECASE & re.UNICODE)
    if guest:
        return [g.strip() for g in re.split(r'\s(?:and(?! the))\s', guest[1])]
    return []


YOUTUBE_ID_MATCHER = re.compile(
    r'^.*(?:(?:youtu\.be\/|v\/|vi\/|u\/\w\/|embed\/)|(?:(?:watch)?\?v(?:i)?=|\&v(?:i)?=))([^#\&\?]+).*')


def extract_youtube_link(post_body):
    soup = bs4.BeautifulSoup(post_body, 'html.parser')
    try:
        youtube_link = json.loads(soup.find('div', class_='video-block')['data-block-json'])['url']
        return youtube_link
    except Exception:
        pass

    try:
        youtube_link = None
        y_links = soup.find_all('a', href=YOUTUBE_ID_MATCHER)
        for yl in y_links:
            m = re.search(YOUTUBE_ID_MATCHER, yl['href'])
            _youtube_link = 'https://www.youtube.com/watch?v=' + m.group(1)
            if youtube_link and _youtube_link != youtube_link:
                logging.warning(f'Multiple youtube links for ep: {y_links}')
                break
            youtube_link = _youtube_link
        return youtube_link or ''
    except Exception:
        return ''


def extract_youtube_id(youtube_link):
    if youtube_link:
        m = YOUTUBE_ID_MATCHER.match(youtube_link)
        return m.group(1) if m else ''
    else:
        return ''


def add_youtube_resources(raw):
    raw_youtube_link = extract_youtube_link(raw['body'])
    youtube_id = extract_youtube_id(raw_youtube_link)

    if youtube_id:
        raw['youtube_id'] = youtube_id
        raw['youtube_link'] = f'https://www.youtube.com/watch?v={youtube_id}'
        raw['youtube_image_link'] = f'https://img.youtube.com/vi/{youtube_id}/mqdefault.jpg'
    else:
        raw['youtube_id'] = None
        raw['youtube_link'] = raw_youtube_link
        raw['youtube_image_link'] = None

    return raw


def fetch_paginated_seq(base_url, route):
    items = []

    while route:
        r = None
        link = furl(base_url + route).add({'format': 'json'})
        try:
            r = requests.get(link)
            res = r.json(encoding='utf-8')

            items += res.get('items', [])
            route = res.get('pagination', {}).get('nextPageUrl', None)
        except Exception:
            logging.error(f'Failed request: {link}')
            logging.error(f'Status code: {r and r.status_code}')
            logging.error(f'Response:\n{r and r.text}')
            raise

    return seq(items)


def match_lists(ings, methods):
    if len(ings) == len(methods):
        return zip([[i] for i in ings], methods)

    return [
        ([i for i in ings if i.text in m.text], m)
        for m in methods
    ]


def extract_recipes(episode):
    recipes = []
    soup = bs4.BeautifulSoup(episode['body'], 'html.parser')

    recipe_locations = soup.find_all(['h1', 'h2', 'h3', 'h4', 'h5'], string=re.compile('Ingredients'))

    for loc in recipe_locations:
        ingredient_lists = loc.parent.find_all(['ul'])

        for iloc in ingredient_lists:
            # TODO: This logic is messy and brittle. Refactor it.

            # Get method name, usually right before the list itself
            method_name = iloc.find_previous_sibling(['p', 'h2', 'h3'])
            if method_name is not None:
                method_name = method_name.get_text().replace('Ingredients', '').strip(string.whitespace + '\u00a0:,')

            if method_name is None or method_name == '':
                # Look for the method name above the recipe instead
                h = iloc.parent.find_next(['h1', 'h2'])
                if h:
                    method_name = h.get_text().strip()
                    method_name = re.sub('^(Method)', '', method_name).strip(string.whitespace + '\u00a0:,')
                else:
                    # No luck, fall back to default
                    method_name = ''

            # Default to the episode name
            if method_name == '':
                method_name = episode['name']

            def flatten(loc):
                out = []
                for i in loc.children:
                    # flatten nested lists
                    if isinstance(i, bs4.element.Tag) and i.select('ul,ol'):
                        for i2 in i.select('ul,ol'):
                            out += flatten(i2)
                    else:
                        out.append(i)
                return out

            # Convert the ul to parsed Ingredients
            try:
                raw_ingredients = flatten(iloc)
                ingredients = [RecipeParser.parse_ingredient(i) for i in raw_ingredients]
            except Exception:
                print("ERROR: Failed to parse ingredients from ep: {0} method: {1} raw_list: {2}".format(
                    episode['name'],
                    method_name,
                    list(iloc.children)
                ))
                raise

            if len(ingredients) == 0:
                print(
                    "WARN: Could not find ingredients for {0} (Episode {1})".format(
                        method_name, episode['name']))

            recipe = {
                'name': method_name,
                'image_link': None,
                'raw_ingredient_list': '\n'.join(i.get_text().strip() for i in raw_ingredients),
                'raw_procedure': '',
                'episode_id': episode['id'],
            }
            recipes.append(recipe)

    return recipes



from flask import jsonify, request, Blueprint, current_app

base_bp = Blueprint('base_bp', __name__, template_folder='templates')


@base_bp.route('/')
def index():
    return jsonify({
        'description': 'Internet Babish DataBase (IBDB) API',
        'routes': sorted(
            set(
                f'{request.base_url[:-1]}{rule}'
                for rule in current_app.url_map.iter_rules()
            )
        )[:-1],
    })
